from functools import wraps
from inspect import getmembers, iscoroutinefunction
from typing import Type, Callable, TYPE_CHECKING

from .. import Client as _Client
from .thread import EventLoopThread

if TYPE_CHECKING:
    import spotify


def _infer_initializer(base: Type, name: str) -> Callable[..., None]:
    """Infer the __init__ to use for a given :class:`typing.Type` base and :class:`str` name."""
    if name in {"HTTPClient", "HTTPUserClient"}:

        def initializer(self: "spotify.HTTPClient", *args, **kwargs) -> None:
            base.__init__(self, *args, **kwargs)
            self.__client_thread__ = kwargs["loop"].__spotify_thread__  # type: ignore

    else:
        assert name != "Client"

        def initializer(self: "spotify.SpotifyBase", client: _Client, *args, **kwargs) -> None:  # type: ignore
            base.__init__(self, client, *args, **kwargs)
            self.__client_thread__ = client.__client_thread__  # type: ignore

    return initializer


def _normalize_coroutine_function(corofunc):
    if isinstance(corofunc, classmethod):

        @classmethod
        @wraps(corofunc)
        def wrapped(cls, client, *args, **kwargs):
            assert isinstance(client, _Client)
            client.__client_thread__.run_coroutine_threadsafe(
                corofunc(cls, client, *args, **kwargs)
            )

    else:

        @wraps(corofunc)
        def wrapped(self, *args, **kwargs):
            return self.__client_thread__.run_coroutine_threadsafe(
                corofunc(self, *args, **kwargs)
            )

    return wrapped


class Synchronous(type):
    """Metaclass used for overloading coroutine functions on models."""

    def __new__(cls, name, bases, dct):
        klass = super().__new__(cls, name, bases, dct)

        base = bases[0]
        name = base.__name__

        if name != "Client":
            # Models and the HTTP classes get their __init__ overloaded.
            initializer = _infer_initializer(base, name)
            setattr(klass, "__init__", initializer)

        for ident, obj in getmembers(base):
            if not iscoroutinefunction(obj):
                continue

            setattr(klass, ident, _normalize_coroutine_function(obj))

        return klass  # type: ignore


class Client(_Client, metaclass=Synchronous):
    def __init__(self, *args, **kwargs):
        thread = EventLoopThread()
        thread.start()

        kwargs["loop"] = thread.loop  # pylint: disable=protected-access

        super().__init__(*args, **kwargs)
        self.__thread = self.__client_thread__ = thread



from .. import _clean_namespace
from . import typing

from .base import AsyncIterable, SpotifyBase, URIBase
from .common import Device, Context, Image
from .artist import Artist
from .track import Track, PlaylistTrack
from .player import Player
from .album import Album
from .library import Library
from .playlist import Playlist
from .user import User

__all__ = (
    "User",
    "Track",
    "PlaylistTrack",
    "Artist",
    "Album",
    "Playlist",
    "Library",
    "Player",
    "Device",
    "Context",
    "Image",
)



from re import S
from typing import Optional

from . import SearchType
from discord.ext import commands


class Track:
    """The base track object. Returns critical track information needed for parsing by Lavalink.
       You can also pass in commands.Context to get a discord.py Context object in your track.
    """

    def __init__(
        self,
        track_id: str,
        info: dict,
        ctx: Optional[commands.Context],
        search_type: SearchType = None,
        spotify: bool = False
    ):
        self.track_id = track_id
        self.info = info
        self.spotify = spotify

        self.title = info.get("title")
        self.author = info.get("author")
        self.length = info.get("length")
        self.ctx = ctx
        self.requester = self.ctx.author if ctx else None
        self.identifier = info.get("identifier")
        self.uri = info.get("uri")
        self.is_stream = info.get("isStream")
        self.is_seekable = info.get("isSeekable")
        self.position = info.get("position")

        if self.spotify:
            self.youtube_result = None
            if search_type:
                self.search_type = search_type

    def __eq__(self, other):
        if not isinstance(other, Track):
            return False

        if self.ctx and other.ctx:
            return other.track_id == self.track_id and other.ctx.message.id == self.ctx.message.id

        return other.track_id == self.track_id

    def __str__(self):
        return self.title

    def __repr__(self):
        return f"<Pomice.track title={self.title!r} uri=<{self.uri!r}> length={self.length}>"


class Playlist:
    """The base playlist object.
       Returns critical playlist information needed for parsing by Lavalink.
       You can also pass in commands.Context to get a discord.py Context object in your tracks.
    """

    def __init__(
        self,
        playlist_info: dict,
        tracks: list,
        ctx: Optional[commands.Context] = None,
        spotify: bool = False,
        thumbnail: Optional[str] = None,
        uri: Optional[str] = None,
    ):
        self.playlist_info = playlist_info
        self.tracks_raw = tracks
        self.spotify = spotify

        self.name = playlist_info.get("name")
        self.selected_track = playlist_info.get("selectedTrack")

        
        self._thumbnail = thumbnail
        self._uri = uri
        
        if self.spotify:
            self.tracks = tracks

        else:
            self.tracks = [
                Track(track_id=track["track"], info=track["info"], ctx=ctx)
                for track in self.tracks_raw
            ]

        self.track_count = len(self.tracks)

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"<Pomice.playlist name={self.name!r} track_count={len(self.tracks)}>"

    @property
    def uri(self) -> Optional[str]:
        """Spotify album/playlist URI, or None if not a Spotify object."""
        return self._uri

    @property
    def thumbnail(self) -> Optional[str]:
        """Spotify album/playlist thumbnail, or None if not a Spotify object."""
        return self._thumbnail



from enum import Enum

class SearchType(Enum):
    """The base class for the different search types for Pomice. 
    This feature is exclusively for the Spotify search feature of Pomice.
    If you are not using this feature, this class is not necessary.

    SearchType.YTSEARCH searches for a Spotify track using regular Youtube, which is best for all scenarios

    SearchType.YTMSEARCH searches for a Spotify track using YouTube Music, which is best for getting audio-only results.

    SearchType.SCSEARCH searches for a Spotify track using SoundCloud, which is an alternative to YouTube or YouTube Music.
    """
    YTSEARCH = 'ytsearch:'
    YTMSEARCH = 'ytmsearch:'
    SCSEARCH = 'scsearch:'
    
    def __str__(self) -> str:
        return self.value




    
        
    


from functools import partial
from typing import Optional, List

from ..oauth import set_required_scopes
from . import AsyncIterable, URIBase, Image, Artist, Track


class Album(URIBase, AsyncIterable):  # pylint: disable=too-many-instance-attributes
    """A Spotify Album.

    Attributes
    ----------
    artists : List[Artist]
        The artists for the album.
    id : str
        The ID of the album.
    name : str
        The name of the album.
    href : str
        The HTTP API URL for the album.
    uri : str
        The URI for the album.
    album_group : str
        ossible values are “album”, “single”, “compilation”, “appears_on”.
        Compare to album_type this field represents relationship between the artist and the album.
    album_type : str
        The type of the album: one of "album" , "single" , or "compilation".
    release_date : str
        The date the album was first released.
    release_date_precision : str
        The precision with which release_date value is known: year, month or day.
    genres : List[str]
        A list of the genres used to classify the album.
    label : str
        The label for the album.
    popularity : int
        The popularity of the album. The value will be between 0 and 100, with 100 being the most popular.
    copyrights : List[Dict]
        The copyright statements of the album.
    markets : List[str]
        The markets in which the album is available: ISO 3166-1 alpha-2 country codes.
    """

    def __init__(self, client, data):
        self.__client = client

        # Simple object attributes.
        self.type = data.pop("album_type", None)
        self.group = data.pop("album_group", None)
        self.artists = [Artist(client, artist) for artist in data.pop("artists", [])]

        self.artist = self.artists[0] if self.artists else None
        self.markets = data.pop("avaliable_markets", None)
        self.url = data.pop("external_urls").get("spotify", None)
        self.id = data.pop("id", None)  # pylint: disable=invalid-name
        self.name = data.pop("name", None)
        self.href = data.pop("href", None)
        self.uri = data.pop("uri", None)
        self.release_date = data.pop("release_date", None)
        self.release_date_precision = data.pop("release_date_precision", None)
        self.images = [Image(**image) for image in data.pop("images", [])]
        self.restrictions = data.pop("restrictions", None)

        # Full object attributes
        self.genres = data.pop("genres", None)
        self.copyrights = data.pop("copyrights", None)
        self.label = data.pop("label", None)
        self.popularity = data.pop("popularity", None)
        self.total_tracks = data.pop("total_tracks", None)

        # AsyncIterable attrs
        self.__aiter_klass__ = Track
        self.__aiter_fetch__ = partial(
            self.__client.http.album_tracks, self.id, limit=50
        )

    def __repr__(self):
        return f"<spotify.Album: {(self.name or self.id or self.uri)!r}>"

    # Public

    @set_required_scopes(None)
    async def get_tracks(
        self, *, limit: Optional[int] = 20, offset: Optional[int] = 0
    ) -> List[Track]:
        """get the albums tracks from spotify.

        Parameters
        ----------
        limit : Optional[int]
            The limit on how many tracks to retrieve for this album (default is 20).
        offset : Optional[int]
            The offset from where the api should start from in the tracks.

        Returns
        -------
        tracks : List[Track]
            The tracks of the artist.
        """
        data = await self.__client.http.album_tracks(
            self.id, limit=limit, offset=offset
        )
        return list(Track(self.__client, item, album=self) for item in data["items"])

    @set_required_scopes(None)
    async def get_all_tracks(
        self, *, market: Optional[str] = "US"
    ) -> List[Track]:  # pylint: disable=unused-argument
        """loads all of the albums tracks, depending on how many the album has this may be a long operation.

        Parameters
        ----------
        market : Optional[str]
            An ISO 3166-1 alpha-2 country code. Provide this parameter if you want to apply Track Relinking.

        Returns
        -------
        tracks : List[:class:`spotify.Track`]
            The tracks of the artist.
        """
        return [track async for track in self]



# pylint: skip-file

from spotify import *
from spotify import __all__, _types, Client
from spotify.utils import clean as _clean_namespace

from . import models
from .models import Client, Synchronous as _Sync


with _clean_namespace(locals(), "name", "base", "Mock"):
    for name, base in _types.items():

        class Mock(base, metaclass=_Sync):  # type: ignore
            __slots__ = {"__client_thread__"}

        Mock.__name__ = base.__name__
        Mock.__qualname__ = base.__qualname__
        Mock.__doc__ = base.__doc__

        locals()[name] = Mock
        setattr(models, name, Mock)

    Client._default_http_client = locals()[
        "HTTPClient"
    ]  # pylint: disable=protected-access



"""Pomice wrapper for Lavalink, made possible by cloudwithax 2021"""

__version__ = "1.0.5"
__title__ = "pomice"
__author__ = "cloudwithax"

import discord

if discord.__version__ != '2.0.0a':
    raise exit("You must have discord.py 2.0 to use this library. Uninstall your current version and install discord.py 2.0 using 'pip install git+https://github.com/Rapptz/discord.py'")

from .enums import SearchType
from .events import *
from .exceptions import *
from .filters import *
from .pool import *
from .objects import *
from .player import Player





__version__ = "0.10.2"
__title__ = "spotify"
__author__ = "mental"
__license__ = "MIT"

from typing import Dict, Type

from .oauth import *
from .utils import clean as _clean_namespace
from .errors import *
from .models import *
from .client import *
from .models import SpotifyBase
from .http import HTTPClient, HTTPUserClient

__all__ = tuple(name for name in locals() if name[0] != "_")

_locals = locals()  # pylint: disable=invalid-name

_types: Dict[str, Type[Union[SpotifyBase, HTTPClient]]]
with _clean_namespace(locals(), "_locals", "_clean_namespace"):
    _types = dict(  # pylint: disable=invalid-name
        (name, _locals[name])
        for name, obj in _locals.items()
        if isinstance(obj, type) and issubclass(obj, SpotifyBase)
    )
    _types["HTTPClient"] = HTTPClient
    _types["HTTPUserClient"] = HTTPUserClient



class Image:
    """An object representing a Spotify image resource.

    Attributes
    ----------
    height : :class:`str`
        The height of the image.
    width : :class:`str`
        The width of the image.
    url : :class:`str`
        The URL of the image.
    """

    __slots__ = ("height", "width", "url")

    def __init__(self, *, height: str, width: str, url: str):
        self.height = height
        self.width = width
        self.url = url

    def __repr__(self):
        return f"<spotify.Image: {self.url!r} (width: {self.width!r}, height: {self.height!r})>"

    def __eq__(self, other):
        return type(self) is type(other) and self.url == other.url


class Context:
    """A Spotify Context.

    Attributes
    ----------
    type : str
        The object type, e.g. “artist”, “playlist”, “album”.
    href : str
        A link to the Web API endpoint providing full details of the track.
    external_urls : str
        External URLs for this context.
    uri : str
        The Spotify URI for the context.
    """

    __slots__ = ("external_urls", "type", "href", "uri")

    def __init__(self, data):
        self.external_urls = data.get("external_urls")
        self.type = data.get("type")

        self.href = data.get("href")
        self.uri = data.get("uri")

    def __repr__(self):
        return f"<spotify.Context: {self.uri!r}>"

    def __eq__(self, other):
        return type(self) is type(other) and self.uri == other.uri


class Device:
    """A Spotify Users device.

    Attributes
    ----------
    id : str
        The device ID
    name : int
        The name of the device.
    type : str
        A Device type, such as “Computer”, “Smartphone” or “Speaker”.
    volume : int
        The current volume in percent. This may be null.
    is_active : bool
        if this device is the currently active device.
    is_restricted : bool
        Whether controlling this device is restricted.
        At present if this is “true” then no Web API commands will be accepted by this device.
    is_private_session : bool
        If this device is currently in a private session.
    """

    __slots__ = (
        "id",
        "name",
        "type",
        "volume",
        "is_active",
        "is_restricted",
        "is_private_session",
    )

    def __init__(self, data):
        self.id = data.get("id")  # pylint: disable=invalid-name
        self.name = data.get("name")
        self.type = data.get("type")

        self.volume = data.get("volume_percent")

        self.is_active = data.get("is_active")
        self.is_restricted = data.get("is_restricted")
        self.is_private_session = data.get("is_private_session")

    def __eq__(self, other):
        return type(self) is type(other) and self.id == other.id

    def __repr__(self):
        return f"<spotify.Device: {(self.name or self.id)!r}>"

    def __str__(self):
        return self.id
